"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const widgets_1 = require("@lumino/widgets");
const dashboard_1 = require("./dashboard");
const clusters_1 = require("./clusters");
/**
 * A widget for hosting Dask dashboard launchers.
 */
class DaskSidebar extends widgets_1.Widget {
    /**
     * Create a new Dask sidebar.
     */
    constructor(options) {
        super();
        this.addClass('dask-DaskSidebar');
        let layout = (this.layout = new widgets_1.PanelLayout());
        // Add the dashboard component/
        this._dashboard = new dashboard_1.DaskDashboardLauncher({
            launchItem: options.launchDashboardItem,
            linkFinder: options.linkFinder
        });
        // A callback that sets the url of the dashboard component.
        const setDashboardUrl = (url) => {
            this._dashboard.input.url = url;
        };
        const injectClientCodeForCluster = options.clientCodeInjector;
        const getClientCodeForCluster = options.clientCodeGetter;
        // Add the cluster manager component.
        this._clusters = new clusters_1.DaskClusterManager({
            registry: options.registry,
            launchClusterId: options.launchClusterId,
            setDashboardUrl,
            injectClientCodeForCluster,
            getClientCodeForCluster
        });
        layout.addWidget(this._dashboard);
        layout.addWidget(this._clusters);
    }
    /**
     * Get the dashboard launcher associated with the sidebar.
     */
    get dashboardLauncher() {
        return this._dashboard;
    }
    /**
     * Get the cluster manager associated with the sidebar.
     */
    get clusterManager() {
        return this._clusters;
    }
}
exports.DaskSidebar = DaskSidebar;
